//
//  main.c
//  osc2_tilde
//
//  Created by Robert Esler on 11/24/15.
//  Copyright © 2015 Robert Esler. All rights reserved.
//

#include "m_pd.h"
#include "math.h"

#define UNITBIT32 1572864.  /* 3*2^19; bit 32 has place value 1 */
#define LOGCOSTABSIZE 9
#define COSTABSIZE (1<<LOGCOSTABSIZE)

#if defined(__FreeBSD__) || defined(__APPLE__) || defined(__FreeBSD_kernel__) \
|| defined(__OpenBSD__)
#include <machine/endian.h>
#endif

#if defined(__linux__) || defined(__CYGWIN__) || defined(__GNU__) || \
defined(ANDROID)
#include <endian.h>
#endif

#ifdef __MINGW32__
#include <sys/param.h>
#endif

#ifdef _MSC_VER
/* _MSVC lacks BYTE_ORDER and LITTLE_ENDIAN */
#define LITTLE_ENDIAN 0x0001
#define BYTE_ORDER LITTLE_ENDIAN
#endif

#if !defined(BYTE_ORDER) || !defined(LITTLE_ENDIAN)
#error No byte order defined
#endif

#if BYTE_ORDER == LITTLE_ENDIAN
# define HIOFFSET 1
# define LOWOFFSET 0
#else
# define HIOFFSET 0    /* word offset to find MSB */
# define LOWOFFSET 1    /* word offset to find LSB */
#endif

union tabfudge
{
    double tf_d;
    int32_t tf_i[2];
};

/* ------------------------ osc2_tilde~ ----------------------------- */

double *cos_table2;

static t_class *osc2_tilde_class;

typedef struct _osc2_tilde
{
    t_object x_obj;
    double x_phase;
    float x_conv;
    float x_f;      /* frequency if scalar */
} t_osc2_tilde;

void *osc2_tilde_new(t_floatarg f)
{
    t_osc2_tilde *x = (t_osc2_tilde *)pd_new(osc2_tilde_class);
    x->x_f = f;
    outlet_new(&x->x_obj, gensym("signal"));
    inlet_new(&x->x_obj, &x->x_obj.ob_pd, &s_float, gensym("ft1"));
    x->x_phase = 0;
    x->x_conv = 0;
    return (x);
}

t_int *osc2_tilde_perform(t_int *w)
{
    t_osc2_tilde *x = (t_osc2_tilde *)(w[1]);
    t_float *in = (t_float *)(w[2]);
    t_float *out = (t_float *)(w[3]);
    int n = (int)(w[4]);
    double *tab = cos_table2, *addr, f1, f2, frac;
    double dphase = x->x_phase + UNITBIT32;
    int normhipart;
    union tabfudge tf;
    float conv = x->x_conv;
    
    tf.tf_d = UNITBIT32;
    normhipart = tf.tf_i[HIOFFSET];
#if 0
    while (n--)
    {
        tf.tf_d = dphase;
        dphase += *in++ * conv;
        addr = tab + (tf.tf_i[HIOFFSET] & (COSTABSIZE-1));
        tf.tf_i[HIOFFSET] = normhipart;
        frac = tf.tf_d - UNITBIT32;
        f1 = addr[0];
        f2 = addr[1];
        *out++ = f1 + frac * (f2 - f1);
    }
#endif
#if 1
    tf.tf_d = dphase;
    dphase += *in++ * conv;
    addr = tab + (tf.tf_i[HIOFFSET] & (COSTABSIZE-1));
    tf.tf_i[HIOFFSET] = normhipart;
    frac = tf.tf_d - UNITBIT32;
    while (--n)
    {
        tf.tf_d = dphase;
        f1 = addr[0];
        dphase += *in++ * conv;
        f2 = addr[1];
        addr = tab + (tf.tf_i[HIOFFSET] & (COSTABSIZE-1));
        tf.tf_i[HIOFFSET] = normhipart;
        *out++ = f1 + frac * (f2 - f1);
        frac = tf.tf_d - UNITBIT32;
    }
    f1 = addr[0];
    f2 = addr[1];
    *out++ = f1 + frac * (f2 - f1);
#endif
    
    tf.tf_d = UNITBIT32 * COSTABSIZE;
    normhipart = tf.tf_i[HIOFFSET];
    tf.tf_d = dphase + (UNITBIT32 * COSTABSIZE - UNITBIT32);
    tf.tf_i[HIOFFSET] = normhipart;
    x->x_phase = tf.tf_d - UNITBIT32 * COSTABSIZE;
    return (w+5);
}

void osc2_tilde_dsp(t_osc2_tilde *x, t_signal **sp)
{
    x->x_conv = COSTABSIZE/sp[0]->s_sr;
    dsp_add(osc2_tilde_perform, 4, x, sp[0]->s_vec, sp[1]->s_vec, sp[0]->s_n);
}

void osc2_tilde_ft1(t_osc2_tilde *x, t_float f)
{
    x->x_phase = COSTABSIZE * f;
}

void cos_maketable(void)
{
    int i;
    double *fp, phase, phsinc = (2. * 3.14159265358979) / COSTABSIZE;
    union tabfudge tf;
    
    if (cos_table2) return;
    cos_table2 = (double *)getbytes(sizeof(double) * (COSTABSIZE+1));
    for (i = COSTABSIZE + 1, fp = cos_table2, phase = 0; i--;
         fp++, phase += phsinc)
        *fp = cos(phase);
    
    /* here we check at startup whether the byte alignment
     is as we declared it.  If not, the code has to be
     recompiled the other way. */
    tf.tf_d = UNITBIT32 + 0.5;
    if ((unsigned)tf.tf_i[LOWOFFSET] != 0x80000000)
        bug("cos~: unexpected machine alignment");
}


void osc2_tilde_setup(void)
{
    osc2_tilde_class = class_new(gensym("osc2~"), (t_newmethod)osc2_tilde_new, 0,
                          sizeof(t_osc2_tilde), 0, A_DEFFLOAT, 0);
    CLASS_MAINSIGNALIN(osc2_tilde_class, t_osc2_tilde, x_f);
    class_addmethod(osc2_tilde_class, (t_method)osc2_tilde_dsp, gensym("dsp"), A_CANT, 0);
    class_addmethod(osc2_tilde_class, (t_method)osc2_tilde_ft1, gensym("ft1"), A_FLOAT, 0);
    
    cos_maketable();
}
